# -*- coding: utf-8 -*-
# Created by crazyX on 2018/7/7

from ojcrawler.control import Controller

import unittest
import os


class Test(unittest.TestCase):

    def setUp(self):
        self.ctl = Controller()
        self.ctl.load_accounts_json(os.path.join(os.getcwd(), 'accounts_sample.json'))

    def test_submit(self):
        pass
        # self.crawler_hdu()
        # self.crawler_poj()
        # self.crawler_cf()

    def test_crawler_hdu(self):
        pid = 1000
        lang = 'g++'
        ac_src = '''
        #include<bits/stdc++.h>
        using namespace std;
        int main()
        {
            int a,b;
            while(cin>>a>>b)cout<<a+b<<endl;
            return 0;
        }
        '''
        wa_src = '''
        #include<bits/stdc++.h>
        using namespace std;
        int main()
        {
            int a,b;
            while(cin>>a>>b)cout<<a-b<<endl;
            return 0;
        }
        '''
        self.ctl.add_task('hdu', ac_src, lang, pid)
        self.ctl.add_task('hdu', wa_src, lang, pid)
        import time
        time.sleep(5)

    def test_crawler_poj(self):
        pid = 1000
        lang = 'g++'
        wa_src = '''
        #include<iostream>
        using namespace std;
        int main()
        {
            int a,b;
            while(cin>>a>>b)cout<<a-b<<endl;
            return 0;
        }
        '''
        ac_src = '''
        #include<iostream>
        using namespace std;
        int main()
        {
            int a,b;
            while(cin>>a>>b)cout<<a+b<<endl;
            return 0;
        }
        '''
        self.ctl.add_task('poj', ac_src, lang, pid)
        self.ctl.add_task('poj', wa_src, lang, pid)
        import time
        time.sleep(20)

    def test_crawler_cf(self):
        pid = '1A'
        lang = 'GNU G++11 5.1.0'
        src = '''
        #include <iostream>
        using namespace std;
        int n,m,a;
        long long x,y;
        int main() {
            cin>>n>>m>>a;
            x=n/a+(n%a==0?0:1);
            y=m/a+(m%a==0?0:1);//sadjiowdqwdw
            cout<<x*y<<endl;
            return 0;
            //fuck you you
        }
        '''
        self.ctl.add_task('codeforces', src, lang, pid)
        import time
        time.sleep(20)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Created by crazyX on 2018/7/14
from bs4 import BeautifulSoup
from robobrowser import RoboBrowser
from socket import timeout
from urllib.error import URLError, HTTPError
from ojcrawler.crawlers.base import OJ
from ojcrawler.crawlers.config import logger, save_image
from ojcrawler.crawlers.config import HTTP_METHOD_TIMEOUT


class Codeforces(OJ):
    def __init__(self, handle, password, image_func=save_image):
        super().__init__(handle, password, image_func)

        self.rb = RoboBrowser(parser='html5lib')
        self.append_html = """
            <!-- MathJax -->
            <script type="text/x-mathjax-config">
            MathJax.Hub.Config({
              tex2jax: {inlineMath: [['$$$','$$$']], displayMath: [['$$$$$$','$$$$$$']]}
            });
            </script>
            <script type="text/javascript" async
                    src="https://assets.codeforces.com/mathjax/MathJax.js?config=TeX-MML-AM_CHTML">
            </script>
            <!-- /MathJax -->
        """
        mathjax_url = 'http://assets.codeforces.com/mathjax/MathJax.js'
        new_mathjax_url = self.image_func(mathjax_url, self.oj_name)
        self.append_html = self.append_html.replace(mathjax_url, new_mathjax_url)

    @property
    def browser(self):
        return self.rb

    @property
    def url_home(self):
        return 'http://codeforces.com/'

    def url_problem(self, cpid):
        # codeforces需要一个cid和一个pid来确定题目
        cid, pid = self.split_pid(cpid)
        return self.url_home + 'problemset/problem/{}/{}'.format(cid, pid)

    @property
    def url_login(self):
        return self.url_home + 'enter/'

    @property
    def url_submit(self):
        return self.url_home + 'problemset/submit/'

    @property
    def url_status(self):
        return self.url_home + 'problemset/status/'

    @property
    def http_headers(self):
        return {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
                          'Ubuntu Chromium/52.0.2743.116 Chrome/52.0.2743.116 Safari/537.36',
            'Origin': "http://codeforces.com",
            'Host': "codeforces.com",
            'Content-Type': 'application/x-www-form-urlencoded',
            'Connection': 'keep-alive',
        }

    @property
    def uncertain_result_status(self):
        # 注意这里不是完全相等而是in
        return ['running', 'in queue']

    def get(self, url):
        try:
            self.browser.open(url, timeout=HTTP_METHOD_TIMEOUT)
            return self.browser.url == url
        except (HTTPError, URLError) as error:
            logger.error('Data not retrieved because %s\nURL: %s', error, url)
            return None
        except timeout:
            logger.error('socket timed out\nURL: %s', url)
            return None

    def post(self, url, data):
        try:
            return self.browser.open(url, 'post', data=data, timeout=HTTP_METHOD_TIMEOUT)
        except (HTTPError, URLError) as error:
            logger.error('Data not retrieved because %s\nURL: %s', error, url)
            return None
        except timeout:
            logger.error('socket timed out\nURL: %s', url)
            return None

    @staticmethod
    def get_languages():
        return {
            'GNU GCC 5.1.0': '10',
            'GNU GCC C11 5.1.0': '43',
            'Clang++17 Diagnostics': '52',
            'GNU G++ 5.1.0': '1',
            'GNU G++11 5.1.0': '42',
            'GNU G++14 6.4.0': '50',
            'GNU G++17 7.3.0': '54',
            'GNU C++17 Diagnostics (DrMemory)': '53',
            'Microsoft Visual C++ 2010': '2',
            'C# Mono 5': '9',
            'D DMD32 v2.079.0': '28',
            'Go 1.8': '32',
            'Haskell GHC 7.8.3': '12',
            'Java 1.8.0_162': '36',
            'Kotlin 1.2': '48',
            'OCaml 4.02.1': '19',
            'Delphi 7': '3',
            'Free Pascal 3': '4',
            'PascalABC.NET 2': '51',
            'Perl 5.20.1': '13',
            'PHP 7.0.12': '6',
            'Python 2.7': '7',
            'Python 3.6': '31',
            'PyPy 2.7 (6.0.0)': '40',
            'PyPy 3.5 (6.0.0)': '41',
            'Ruby 2.0.0p645': '8',
            'Rust 1.26': '49',
            'Scala 2.12': '20',
            'JavaScript V8 4.8.0': '34',
            'Node.js 6.9.1': '55',
            'ActiveTcl 8.5': '14',
            'Io-2008-01-07 (Win32)': '15',
            'Pike 7.8': '17',
            'Befunge': '18',
            'OpenCobol 1.0': '22',
            'Factor': '25',
            'Secret_171': '26',
            'Roco': '27',
            'Ada GNAT 4': '33',
            'Mysterious Language': '38',
            'FALSE': '39',
            'Picat 0.9': '44',
            'GNU C++11 5 ZIP': '45',
            'Java 8 ZIP': '46',
            'J': '47',
            'Microsoft Q#': '56',
        }

    def login(self):
        try:
            self.browser.open(self.url_login, timeout=HTTP_METHOD_TIMEOUT)
            enter_form = self.browser.get_form('enterForm')
            enter_form['handleOrEmail'] = self.handle
            enter_form['password'] = self.password
            self.browser.submit_form(enter_form)
            # checks = list(map(lambda x: x.getText()[1:].strip(),
            #                   self.browser.select('div.caption.titled')))
            if self.browser.url == self.url_home:
                return True, ''
            elif self.browser.url == self.url_login:
                # <span class="error for__password">
                soup = BeautifulSoup(self.browser.response.content, 'html5lib')
                return False, soup.find('span', {'class': 'error for__password'}).text
            else:
                return False, '登陆：未知错误'

        except (HTTPError, URLError) as error:
            return False, '登陆：HTTPError/URLError - ' + str(error)
        except timeout:
            return False, '登陆：http方法超时'

    def is_login(self):
        self.browser.open(self.url_home)
        checks = list(map(lambda x: x.getText()[1:].strip(),
                          self.browser.select('div.caption.titled')))
        return self.handle in checks

    def replace_image(self, html):
        # <img class="tex-graphics" src="/predownloaded/e6/1a/e61a341f8b20d0e351ba708597ce43b451f51e25.png"
        # style="max-width: 100.0%;max-height: 100.0%;">
        pos = html.find('<img')
        if pos == -1:
            return html
        src_pos = html[pos:].find('src=') + pos
        stp = src_pos + 5
        edp = html[stp:].find('"') + stp
        image_url = self.url_home[:-1] + html[stp:edp]
        saved_url = self.image_func(image_url, self.oj_name)
        return html[:stp] + saved_url + self.replace_image(html[edp:])

    def get_problem(self, cpid):
        ret = self.get(self.url_problem(cpid))
        cid, pid = self.split_pid(cpid)
        if ret:
            soup = BeautifulSoup(self.browser.response.content, 'html5lib')
            header = soup.find('div', {'class': 'header'})
            title = header.find('div', {'class': 'title'}).text[len(pid) + 1:].strip()

            interactive = soup.find('span', {'class': 'tex-font-style-bf'})
            problem_type = 'interactive' \
                if (interactive and interactive.text == 'This is an interactive problem.') \
                else 'special judge'

            origin = self.url_problem(cpid)

            limits = {
                'default': (int(float(header.find('div', {'class': 'time-limit'}).contents[1].split(' ')[0]) * 1000),
                            int(header.find('div', {'class': 'memory-limit'}).contents[1].split(' ')[0]) * 1024),
            }

            samples = {}

            descriptions = []
            category = ''
            tags = []

            append_html = self.append_html

            htmls = header.find_next_siblings('div')
            for item in htmls:
                temp = item.find('div', {'class': 'section-title'})
                sub_title = temp.text if temp else ''
                if sub_title == '':
                    sub_content = str(item)
                else:
                    sub_content = ''.join([str(item) for item in temp.find_next_siblings()])
                if sub_title == 'Example':
                    continue
                descriptions.append(
                    (sub_title, sub_content)
                )

            inputs = soup.find_all('div', {'class': 'input'})
            outputs = soup.find_all('div', {'class': 'output'})
            assert len(inputs) == len(outputs)
            n = len(inputs)
            for i in range(n):
                input_html = inputs[i].find('pre')
                for br in input_html.find_all('br'):
                    br.replace_with("\n")
                output_html = outputs[i].find('pre')
                for br in output_html.find_all('br'):
                    br.replace_with("\n")
                samples[i + 1] = (input_html.text, output_html.text)

            category = soup.find('a', {'style': 'color: black'}).text
            tag_htmls = soup.find_all('span', {'class': 'tag-box'})
            for tag_html in tag_htmls:
                tags.append(tag_html.text.strip())

            compatible_data = {}
            for key in self.compatible_problem_fields:
                compatible_data[key] = eval(key)

            for tag in tags:
                if tag.startswith('*'):
                    try:
                        val = int(tag[1:])
                        compatible_data['difficult_number'] = val
                    except ValueError:
                        pass
            return True, compatible_data

        elif ret is None:
            return False, '获取题目：http方法错误，请检查网络后重试'
        else:
            return False, '获取题目：不存在的题目'

    @staticmethod
    def split_pid(cid: str):
        """
        将cf的cid拆分成可用的形式，到第一个非数字字符为止
        :param cid: '123A', '342c2'
        :return: cid, pid
        """
        n = len(cid)
        for i in range(n):
            if not cid[i].isdigit():
                return cid[:i], cid[i:]
        return cid, ''

    def submit_code(self, source, lang, cpid):
        if not self.is_login():
            success, info = self.login()
            if not success:
                return False, info
        cid, pid = self.split_pid(cpid)

        problem_id = '{}{}'.format(cid, pid).upper()

        ret = self.get(self.url_submit)
        if ret:
            submit_form = self.browser.get_form(class_='submit-form')
            submit_form['submittedProblemCode'] = problem_id
            submit_form['source'] = source

            # 注意这儿的语言不是全大写
            submit_form['programTypeId'] = self.get_languages()[lang]
            self.browser.submit_form(submit_form)
            if self.browser.url == self.url_status[:-1]:
                ok, info = self.get_result()
                return (True, info['rid']) if ok else (False, '提交代码（获取提交id）：' + info)
            elif self.url_submit in self.browser.url:
                soup = BeautifulSoup(self.browser.response.content, 'html5lib')
                info = soup.find('span', {'class': 'error for__source'}).text
                return False, info
            else:
                return False, '提交代码：未知错误1'
        elif ret is None:
            return False, '提交代码：http方法错误，请检查网络后重试'
        else:
            return False, '提交代码：未知错误2'

    def get_result(self):
        # 只找最新的
        url = 'http://codeforces.com/submissions/{}'.format(self.handle)
        ret = self.get(url)
        if ret:
            soup = BeautifulSoup(self.browser.response.content, 'html5lib')
            table = soup.find('table', {'class': 'status-frame-datatable'})
            trs = table.find_all('tr')
            if len(trs) <= 1:
                return False, '没有结果'
            data = {
                'rid': trs[1].contents[1].text.strip(),
                'status': trs[1].contents[11].text.strip(),
                'time': trs[1].contents[13].text.strip().split('\xa0')[0],
                'memory': trs[1].contents[15].text.strip().split('\xa0')[0],
                'ce_info': '',
            }
            if data['status'] == 'Compilation error':
                ret, info = self.get_compile_error_info(data['rid'])
                data['ce_info'] = info if ret else ''
            return True, data

        elif ret is None:
            return False, '获取结果：http方法错误，请检查网络后重试'
        else:
            return False, '获取结果：未知错误2'

    def get_result_by_rid(self, rid):
        ret, dat = self.get_result()
        if ret:
            if dat['rid'] != rid:
                info = 'rid not matched! ({}!={}), CF不支持使用rid寻找提交结果'.format(rid, dat['rid'])
                logger.error(info)
                return False, info
            else:
                return ret, dat
        else:
            return ret, dat

    def get_compile_error_info(self, rid):
        ret = self.get(self.url_status)
        if ret:
            soup = BeautifulSoup(self.browser.response.content, 'html5lib')
            csrf = soup.find('meta', {'name': 'X-Csrf-Token'}).attrs['content']
            url = 'http://codeforces.com/data/judgeProtocol'
            data = {'submissionId': rid, 'csrf_token': csrf}
            self.post(url, data)
            soup = BeautifulSoup(self.browser.response.content, 'html5lib')
            info = eval(soup.text)
            return (True, info) if info else (False, '')
        elif ret is None:
            return False, '获取编译错误信息：http方法错误，请检查网络后重试'
        else:
            return False, '获取编译错误信息：未知错误2'



# -*- coding: utf-8 -*-
# Created by crazyX on 2018/7/8
from socket import timeout
from urllib.error import URLError, HTTPError
from ojcrawler.crawlers.config import logger
from ojcrawler.crawlers.config import HTTP_METHOD_TIMEOUT


class OJ(object):
    # 每一个账号同一时间只考虑交一道题目，这样可以有效避免查封，且方便处理
    # image_func 用来做网页中图片的url替换
    def __init__(self, handle, password, image_func):
        self.handle = handle
        self.password = password
        self.image_func = image_func

    def __str__(self):
        return "{}({})".format(self.oj_name, self.handle)

    @property
    def oj_name(self):
        return self.__class__.__name__.lower()

    # 以下为基础属性
    @property
    def browser(self):
        raise NotImplementedError

    @property
    def url_home(self):
        raise NotImplementedError

    def url_problem(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def url_login(self):
        raise NotImplementedError

    @property
    def url_submit(self):
        raise NotImplementedError

    @property
    def url_status(self):
        raise NotImplementedError

    @property
    def http_headers(self):
        raise NotImplementedError

    @property
    def uncertain_result_status(self):
        raise NotImplementedError

    @property
    def compatible_problem_fields(self):
        # title 为标题，字符串
        # time limit 数字，单位为 ms
        # memory limit 数字，单位为 kb
        # problem_type为字符串，表示题目类型，默认为'regular', 可选为'special judge'，'interactive'等
        # origin 为题目链接字符串

        # samples_input\output为list：
        # 注意某些samples可能没有输入或者没有输出

        # descriptions为所有描述，hint, source, 等内容的二元组有序列表，内容都为html源码，并替换了其中的image路径为本地路径
        # [(sub_title1, html1), (sub_title2, html2), ...]

        # 需要额外考虑一下针对不同语言的不同的time limit和memory limit
        # time_limit = {
        #   'default': 1000,
        #   'java': 3000,
        # }
        # memory_limit = {
        #   'default': 65536,
        # }

        # Description, Input, Output, Samples, Hint, Source,

        # 增加两个字段，一个是category，一个是tags
        # 对于poj，将其source的text转为category；tags初始空，后期手动标注；
        # 对于hdu，将其source的text转为category；tags初始空，后期手动标注；
        # 对于codeforces，将其对应比赛名转为category；tags为problem tags；

        # category默认为空字符串，且每个题目最多对应到一个确定的category
        # 每个题目可以对应到多个tag, tags类型为list

        # 增加一个append_html项，可能有依赖的html样式或者mathjax配置等等

        return ['title', 'problem_type', 'origin',
                'limits',
                'samples',
                # 'time_limit', 'memory_limit',
                # 'samples_input', 'samples_output',
                'descriptions',
                'category',
                'tags',
                'append_html',
                ]

    # 以下为基础函数
    def get(self, url):
        try:
            return self.browser.open(url, timeout=HTTP_METHOD_TIMEOUT)
        except (HTTPError, URLError) as error:
            logger.error('Data not retrieved because %s\nURL: %s', error, url)
            return None
        except timeout:
            logger.error('socket timed out\nURL: %s', url)
            return None

    def post(self, url, data):
        raise NotImplementedError

    @staticmethod
    def http_status_code(response):
        return response.status if response else None

    def ping(self):
        # 5s是否能访问主页
        response = self.get(self.url_home)
        return self.http_status_code(response) == 200

    @staticmethod
    def get_languages():
        # 获取语言列表
        # example:
        # LANGUAGE = {
        #     'G++': '1',
        #     'G++11': '42',
        #     'G++14': '50',
        #     'GCC': '10',
        #     'JAVA': '36',
        #     'PYTHON2': '7',
        #     'PYTHON3': '31',
        # }
        raise NotImplementedError

    def login(self):
        raise NotImplementedError

    def is_login(self):
        raise NotImplementedError

    def replace_image(self, html):
        raise NotImplementedError

    def get_problem(self, *args, **kwargs):
        raise NotImplementedError

    def submit_code(self, *args, **kwargs):
        # 返回 run id
        raise NotImplementedError

    def get_result(self):
        # 只需要获取最近一次提交的结果
        # 如果遇到了什么异常，考虑直接重新提交
        raise NotImplementedError

    def get_result_by_rid(self, rid):
        # 这个不一定每个系统都能实现
        pass

    def get_compile_error_info(self, rid):
        # 这个不一定每个系统都能实现
        pass



from ojcrawler.crawlers.config import *
from ojcrawler.crawlers import supports
from queue import Queue
import threading
from time import sleep
import json


def sample_sync_func(data, *args):
    # 多余的对应参数应该在add_task的时候按顺序传入
    # data = {
    #     'status': '各oj对应的状态字符串',
    #     'established': True,  # False, 表明是否是确定的状态，如果是，应该还有额外的信息
    #     'rid': trs[1].contents[0].text.strip(),
    #     'status': trs[1].contents[2].text.strip(),
    #     'time': trs[1].contents[4].text.strip(),
    #     'memory': trs[1].contents[5].text.strip(),
    # }
    json_data = json.dumps(data)
    logger.info("data: " + json_data)


def sample_save_image(image_url, oj_name):
    # 传入一个图片的地址，返回新的地址
    # oj_name 会传入oj自身的名字，方便用来分类
    # 1. 可以将图片保存到本地然后返回静态服务器的地址
    # 2. 可以上传到某图云然后返回图云的地址
    # 3. 也可以直接返回源oj的地址，这样如果不能访问外网就存在风险
    return image_url


class SingletonOJHandle(type):
    # 这个metaclass保证第一属性和第二属性的元组单例
    _map_instance = {}

    def __call__(cls, a, b, *args, **kwargs):
        if (cls, a, b) not in cls._map_instance:
            cls._map_instance[(cls, a, b)] = super().__call__(a, b, *args, **kwargs)
        else:
            logger.warning('you are create a exist oj-handle({}-{}), '
                           'return mapped instead and this will update other attribute.'.format(a, b))
        return cls._map_instance[(cls, a, b)]


class Worker(threading.Thread, metaclass=SingletonOJHandle):
    # 需要保证同一个oj，同一个handle为单例
    def __init__(self, oj_name, handle, password, queue: Queue, image_func, sync_func):
        super().__init__(name='{}-{}'.format(oj_name, handle))
        self.oj = supports[oj_name](handle, password, image_func)
        self.queue = queue
        self.sync_func = sync_func

        # 线程控制相关
        self.__flag = threading.Event()  # 用于暂停线程的标识
        self.__flag.set()  # 设置为True
        self.__running = threading.Event()  # 用于停止线程的标识
        self.__running.set()  # 将running设置为True

    def pause(self):
        self.__flag.clear()  # 设置为False, 让线程阻塞

    def resume(self):
        self.__flag.set()  # 设置为True, 让线程停止阻塞

    def stop(self):
        # # stop的时候存在一个问题，由于中间还需要从queue中获取数据，如果queue已经为空，run函数
        # # 会阻塞在queue.get()函数处，永远无法退出，所以控制器退出的时候需要往queue中push Worker
        # # 个数的None
        #
        # self.__flag.set()  # 将线程从暂停状态恢复, 如何已经暂停的话
        self.__running.clear()  # 设置为False

    # 队列中的信息：
    # pid, source, lang, *args
    # *args, **kwargs和sync_func中保持一致
    def run(self):
        while self.__running.isSet():
            # self.__flag.wait()  # 为True时立即返回, 为False时阻塞直到内部的标识位为True后返回
            item = self.queue.get()
            if item:
                # 注意处理cf的pid
                # cf的pid应该为 123A 之类的形式
                # 内部会将其拆分
                source = item[0]
                lang = item[1]
                pid = item[2]
                args = item[3:]
                success, dat = self.oj.submit_code(source, lang, pid)
                if not success:
                    logger.warning('{} - {}'.format(self.oj.oj_name, dat))
                    # 注意如果处理失败，应该再次放入队列
                    # self.queue.put(item)
                    self.sync_func({'status': 'submit failed', 'established': True}, args)
                    continue

                self.sync_func({'status': 'submitted', 'established': False}, args)
                pre_status = 'submitted'
                cnt = 0
                fetch_success = False
                while cnt < RESULT_COUNT:
                    sleep(RESULT_INTERVAL)
                    success, info = self.oj.get_result_by_rid(dat)
                    if success:
                        status = info['status']
                        if status != pre_status:
                            # 注意codeforces
                            established = True
                            for uncertain_status in self.oj.uncertain_result_status:
                                if str(status).lower() in uncertain_status:
                                    established = False
                            if not established:
                                info['established'] = False
                                self.sync_func(info, args)
                                pre_status = status
                            else:
                                info['established'] = True
                                self.sync_func(info, args)
                                fetch_success = True
                                break
                    cnt = cnt + 1

                if not fetch_success:
                    self.sync_func({'status': 'fetch failed', 'established': False}, args)
            self.queue.task_done()



# -*- coding: utf-8 -*-
# Created by crazyX on 2018/7/7
from ojcrawler.crawlers.poj import POJ
from ojcrawler.crawlers.hdu import HDU
from ojcrawler.crawlers.codeforces import Codeforces

supports = {
    'poj': POJ,
    'hdu': HDU,
    'codeforces': Codeforces,
}



import inspect


def sample_sync_func(status, *args, **kwargs):
    pass


class Controller(object):
    # 一次性根据配置的账号数量，初始化对应oj对应数量的Controller
    # 在外部做负载均衡

    def __init__(self, oj_name: str, sync_func=sample_sync_func):
        self.oj = oj_name
        print(inspect.getfullargspec(sync_func))
        args = inspect.getfullargspec(sync_func)[0]
        if len(args) < 1 or args[0] != 'status':
            raise ValueError(
                'sync_func的第一个参数必须为status而不是{}, sample: sync_func(status, *args, **kwargs)'.format(args[0]))
        self.func = sync_func

    def run(self, *args, **kwargs):
        self.func('test', *args, **kwargs)


def sync_fun(status, ip, port):
    print('status:', status)
    print('ip:', ip)
    print('port:', port)


if __name__ == '__main__':
    c = Controller('x')
    c.run()



# -*- coding: utf-8 -*-
# Created by crazyX on 2018/7/12
from ojcrawler.crawlers import supports
import inspect
import json
from queue import Queue
from ojcrawler.utils import sample_save_image, sample_sync_func, Worker


class Controller(object):

    # 😝不同OJ爬虫的同步状态函数和替换图片url函数已经被抽象为统一的函数
    def __init__(self, sync_func=sample_sync_func, image_func=sample_save_image):
        # 这个函数用来同步状态，必须为sync_func(status, *args, **kwargs) 形式xw
        args = inspect.getfullargspec(sync_func)[0]
        if len(args) < 1 or args[0] != 'data':
            raise ValueError('sync_func的第一个参数必须为data而不是{}, '
                             'sample: sync_func(data, *args, **kwargs)'.format(args[0]))

        args = inspect.getfullargspec(image_func)[0]
        if len(args) != 2:
            raise ValueError('image_func必须为两个参数')
        if args[0] != 'image_url' or args[1] != 'oj_name':
            raise ValueError('image_func的两个参数必须为image_url({})和oj_name({}), '
                             'sample: sample_save_image(image_url, oj_name)'.format(args[0], args[1]))

        self.sync_func = sync_func
        self.image_func = image_func

        self.queues = {}
        # 一个oj可能对应多个worker，{'poj': [instance1, instance2], 'hdu': [instance1]}
        self.workers = {}

        self.static_supports = {}

        for key in supports.keys():
            self.queues[key] = Queue()
            self.workers[key] = []
            self.static_supports[key] = supports[key]('static', 'static', image_func)

    def __del__(self):
        print('正在停止workers')
        self.stop()
        print('停止成功')

    @staticmethod
    def supports():
        return supports.keys()

    def _add_account(self, oj_name, handle, password):
        # 同一个oj重复handle只会采用第一个的配置
        worker = Worker(oj_name, handle, password, self.queues[oj_name], self.image_func, self.sync_func)
        # 可能是已经存在的实例
        if worker not in self.workers[oj_name]:
            self.workers[oj_name].append(worker)
            worker.start()

    def load_accounts_json(self, json_path):
        with open(json_path) as fp:
            json_data = json.load(fp)
        accounts = []
        for oj_name in json_data:
            if oj_name not in supports.keys():
                raise NotImplementedError('oj_name only supports: {}'.format(str(supports.keys())))
            for item in json_data[oj_name]:
                accounts.append((oj_name, item['handle'], item['password']))
        self.init_accounts(accounts)

    def init_accounts(self, accounts):
        # 初始化account信息，注意不能用重复的信息初始化
        # 注意会清空之前的账号信息
        for oj_name, handle, password in accounts:
            if oj_name not in supports.keys():
                raise NotImplementedError('oj_name only supports: {}'.format(str(supports.keys())))

        # 先停止所有的worker
        self.stop()
        # 创建对应的队列集和工作者集
        for key in supports.keys():
            self.queues[key] = Queue()
            self.workers[key] = []
        for oj_name, handle, password in accounts:
            self._add_account(oj_name, handle, password)
        return True

    def add_task(self, oj_name, source, lang, pid, *args):
        if oj_name not in supports.keys():
            raise NotImplementedError('oj_name only supports: {}'.format(str(supports.keys())))
        self.queues[oj_name].put((source, lang, pid, *args))

    def start(self):
        if not self.workers:
            raise RuntimeError('you should init accounts first.')
        for key in self.workers:
            for worker in self.workers[key]:
                worker.setDaemon(True)
                worker.start()

    def pause(self):
        for key in self.workers:
            for worker in self.workers[key]:
                worker.pause()

    def stop(self):

        for key in self.workers:
            for worker in self.workers[key]:
                assert type(worker) == Worker
                worker.stop()

        # for key in self.queues.keys():
        #     cnt = len(self.workers[key])
        #     for i in range(cnt):
        #         self.queues[key].put(None)
        #
        # for key in self.workers:
        #     for worker in self.workers[key]:
        #         worker.join()

        # 清空worker和队列内存
        for queue in self.queues.values():
            with queue.mutex:
                queue.queue.clear()
            del queue
        for key in self.workers:
            for worker in self.workers[key]:
                del worker
        self.queues = {}
        self.workers = {}

    def get_languages(self, oj_name):
        if oj_name not in supports.keys():
            raise NotImplementedError('oj_name only supports: {}'.format(str(supports.keys())))
        return self.static_supports[oj_name].get_languages()

    @staticmethod
    def get_basic_language(oj_name):
        if oj_name not in supports.keys():
            raise NotImplementedError('oj_name only supports: {}'.format(str(supports.keys())))
        # 只考虑三种最基础的语言，用来在比赛当中避免选手根据源语言判断OJ来源
        # c, c++, java
        if oj_name == 'poj':
            return {
                'c': 'GCC',
                'c++': 'G++',
                'c++11': None,
                'java': 'JAVA',
            }
        elif oj_name == 'hdu':
            return {
                'c': 'GCC',
                'c++': 'G++',
                'c++11': 'G++',
                'java': 'JAVA',
            }

        elif oj_name == 'codeforces':
            return {
                'c': 'GNU GCC C11 5.1.0',
                'c++': 'GNU G++11 5.1.0',
                'c++11': 'GNU G++11 5.1.0',
                'java': 'Java 1.8.0_162',
            }

    def get_problem(self, oj_name, pid):
        if oj_name not in supports.keys():
            raise NotImplementedError('oj_name only supports: {}'.format(str(supports.keys())))
        return self.static_supports[oj_name].get_problem(pid)



from logging.handlers import RotatingFileHandler
import os
import urllib
import urllib.request
import logging

logFile = 'crawler.log'
my_handler = RotatingFileHandler(logFile, mode='a', maxBytes=5 * 1024 * 1024,
                                 backupCount=2, encoding=None, delay=0)

formatter = logging.Formatter('%(name)s: %(asctime)s [%(module)s:%(funcName)s] [%(levelname)s]- %(message)s')
my_handler.setFormatter(formatter)

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s [%(threadName)s:%(thread)d] [%(name)s:%(lineno)d]'
                           ' [%(module)s:%(funcName)s] [%(levelname)s]- %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename='judge.log',
                    filemode='w',
                    )

logging.getLogger('').addHandler(my_handler)


DEBUG = True

if DEBUG:
    # define a Handler which writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    # set a format which is simpler for console use

    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger('').addHandler(console)

    # Now, we can log to the root logger, or any other logger. First the root...


logger = logging

# 超时秒数
HTTP_METHOD_TIMEOUT = os.getenv('HTTP_METHOD_TIMEOUT', 10)

# 获取结果次数
RESULT_COUNT = os.getenv('RESULT_COUNT', 20)

# 每两次获取结果之间间隔 / s
RESULT_INTERVAL = os.getenv('RESULT_INTERVAL', 1)

# 静态目录
STATIC_OJ_ROOT = os.getenv('STATIC_OJ_ROOT', '/home/')

# 静态url
STATIC_OJ_URL = os.getenv('STATIC_OJ_URL', 'localhost:8000/statics/')


def save_image(image_url, oj_name):
    # 保存静态文件，并返回新的url

    file_name = image_url.split('/')[-1]

    image_folder = os.path.join(STATIC_OJ_ROOT, oj_name)
    if not os.path.exists(image_folder):
        os.makedirs(image_folder)
    path = os.path.join(image_folder, file_name)
    if os.path.exists(path):
        return STATIC_OJ_URL + oj_name + '/' + file_name

    req = urllib.request.Request(image_url)
    resp = urllib.request.urlopen(req)
    data = resp.read()
    with open(path, 'wb') as fp:
        fp.write(data)
    return STATIC_OJ_URL + oj_name + '/' + file_name



from setuptools import setup, find_packages

with open("README.md", encoding='utf-8', mode="r") as fh:
    long_description = fh.read()

setup(
    name='ojcrawler',
    version='1.1.10',
    packages=find_packages(),
    url='https://github.com/LETTersOnline/ojcrawler',
    license='MIT',
    author='crazyX',
    author_email='xu_jingwei@outlook.com',
    description='crawler of some online judge system',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['psutil', 'requests', 'robobrowser', 'beautifulsoup4', 'html5lib'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# -*- coding: utf-8 -*-
# Created by crazyX on 2018/7/13
from socket import timeout
from http import cookiejar
from bs4 import BeautifulSoup
from urllib import request, parse
from urllib.error import URLError, HTTPError

from ojcrawler.crawlers.base import OJ
from ojcrawler.crawlers.config import logger, save_image
from ojcrawler.crawlers.config import HTTP_METHOD_TIMEOUT


class HDU(OJ):
    def __init__(self, handle, password, image_func=save_image):
        super().__init__(handle, password, image_func)

        # 声明一个CookieJar对象实例来保存cookie
        cookie = cookiejar.CookieJar()
        # 利用urllib2库的HTTPCookieProcessor对象来创建cookie处理器
        handler = request.HTTPCookieProcessor(cookie)
        # 通过handler来构建opener
        self.opener = request.build_opener(handler)
        # 此处的open方法同urllib2的urlopen方法，也可以传入request

    @property
    def browser(self):
        return self.opener

    @property
    def url_home(self):
        return 'http://acm.hdu.edu.cn/'

    def url_problem(self, pid):
        return self.url_home + 'showproblem.php?pid={}'.format(pid)

    @property
    def url_login(self):
        return self.url_home + 'userloginex.php?action=login'

    @property
    def url_submit(self):
        return self.url_home + 'submit.php?action=submit'

    @property
    def url_status(self):
        return self.url_home + 'status.php?'

    @property
    def http_headers(self):
        return {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
                          'Ubuntu Chromium/52.0.2743.116 Chrome/52.0.2743.116 Safari/537.36',
            'Origin': "http://acm.hdu.edu.cn",
            'Host': "acm.hdu.edu.cn",
            'Content-Type': 'application/x-www-form-urlencoded',
            'Connection': 'keep-alive',
        }

    @property
    def uncertain_result_status(self):
        # lower
        return ['queuing', 'compiling', 'running']

    def post(self, url, data):
        post_data = parse.urlencode(data).encode()
        req = request.Request(url, post_data, self.http_headers)
        try:
            return self.opener.open(req, timeout=HTTP_METHOD_TIMEOUT)
        except (HTTPError, URLError) as error:
            logger.error('Data not retrieved because %s\nURL: %s', error, url)
        except timeout:
            logger.error('socket timed out\nURL: %s', url)

    @staticmethod
    def get_languages():
        # hdu支持语言不太可能发生变化
        return {
            'G++': '0',
            'GCC': '1',
            'C++': '2',
            'C': '3',
            'PASCAL': '4',
            'JAVA': '5',
            'C#': '6',
        }

    def login(self):
        data = dict(
            username=self.handle,
            userpass=self.password,
            login='Sign In'
        )
        ret = self.post(self.url_login, data)
        if ret:
            html = ret.read().decode('gbk')
            if html.find('signout') > 0:
                return True, ''
            else:
                return False, '账号密码不匹配'
        else:
            return False, '登陆：http方法错误，请检查网络后重试'

    def is_login(self):
        ret = self.get(self.url_home)
        if ret:
            html = ret.read().decode('gbk')
            return True if html.find('<img alt="Author"') > 0 else False
        else:
            return False

    def replace_image(self, html):
        # http://acm.hdu.edu.cn/showproblem.php?pid=3987
        # http://acm.hdu.edu.cn/showproblem.php?pid=1033
        pos = html.find('<img')
        if pos == -1:
            return html
        src_pos = html[pos:].find('src=')
        stp = pos + src_pos + 5
        edp = html[stp + 1:].find('"') + stp
        url = html[stp: edp + 1]
        image_url = self.url_home + url[url.find('/data/') + 1:]
        saved_url = self.image_func(image_url, self.oj_name)

        return html[:stp] + saved_url + self.replace_image(html[edp + 1:])

    def get_problem(self, pid):
        ret = self.get(self.url_problem(pid))
        if ret:
            html = ret.read()
            # 不需要decode，交给BeautifulSoup处理
            soup = BeautifulSoup(html, 'html5lib')
            msg_png = soup.find('img', {'src': '/images/msg.png'})
            if msg_png:
                return False, soup.find('td', {'valign': 'middle'}).find('div').text
            else:
                title = soup.find('h1', {'style': 'color:#1A5CC8'}).text

                spj = soup.find('font', {'color': 'red'})
                problem_type = 'special judge' if (spj and 'Special Judge' == spj.text) else 'regular'

                origin = self.url_problem(pid)

                limits = soup.find('span', {'style': 'font-family:Arial;font-size:12px;font-weight:bold;color:green'})
                limits_list = limits.contents[0].split(' ')
                time_java, time_default = [int(x) for x in limits_list[2].split('/')]
                memory_java, memory_default = [int(x) for x in limits_list[6].split('/')]
                ac_submit = limits.contents[2].replace('\xa0', ' ').split(' ')

                limits = {
                    'default': (time_default, memory_default),
                    'java': (time_java, memory_java),
                }

                samples_input = []
                samples_output = []
                samples = {}

                descriptions = []
                category = ''
                tags = []
                append_html = ''

                # 题面
                panel_titles = soup.find_all('div', {'class': 'panel_title'})
                panel_contents = soup.find_all('div', {'class': 'panel_content'})
                assert len(panel_titles) == len(panel_contents)
                n = len(panel_titles)
                for i in range(n):
                    if panel_titles[i].text == 'Sample Input':
                        samples_input.append(panel_contents[i].text)
                    elif panel_titles[i].text == 'Sample Output':
                        samples_output.append(panel_contents[i].text)
                    elif panel_titles[i].text == 'Source':
                        category = panel_contents[i].text
                    else:
                        descriptions.append(
                            (panel_titles[i].text,
                             self.replace_image(str(panel_contents[i]))
                             )
                        )
                assert len(samples_input) == len(samples_output) or len(samples_input) == 0
                n = len(samples_output)
                if len(samples_input) != 0:
                    for i in range(n):
                        samples[i + 1] = (samples_input[i], samples_output[i])
                else:
                    for i in range(n):
                        samples[i + 1] = ("", samples_output[i])

                compatible_data = {}
                for key in self.compatible_problem_fields:
                    compatible_data[key] = eval(key)

                # 增加题目提交数和ac数
                try:
                    compatible_data['accepted_number'] = int(ac_submit[-1])
                    compatible_data['submitted_number'] = int(ac_submit[2])
                except Exception as e:
                    logger.warning("获取submit和ac数失败({})：".format(pid) + str(e))

                return True, compatible_data
        else:
            return False, '获取题目：http方法错误，请检查网络后重试'

    def submit_code(self, source, lang, pid):
        if not self.is_login():
            success, info = self.login()
            if not success:
                return False, info
        data = dict(
            problemid=pid,
            language=self.get_languages()[lang.upper()],
            usercode=source,
            check='0',
        )
        ret = self.post(self.url_submit, data)
        if ret:
            # 不直接用重定向页面是因为，并发高的时候，第一个提交并不一定是自己的
            if ret.url == self.url_status[:-1]:
                ok, info = self.get_result()
                return (True, info['rid']) if ok else (False, '提交代码（获取提交id）：' + info)
            elif ret.url == self.url_submit:
                html = ret.read()
                soup = BeautifulSoup(html, 'html5lib')
                return False, soup.find('div', {'style': 'color:red; font-size:12px'}).text
            else:
                return False, '提交代码：未知错误'
        else:
            return False, '提交代码：http方法错误，请检查网络后重试'

    def _get_result(self, url_result):
        # 获取url_result下的第一个结果
        ret = self.get(url_result)
        if ret:
            html = ret.read()
            soup = BeautifulSoup(html, 'html5lib')
            table = soup.find('table', {'class': 'table_text'})
            trs = table.find_all('tr')
            if len(trs) <= 1:
                return False, '没有结果'
            else:
                # 第一行为header
                data = {
                    'rid': trs[1].contents[0].text.strip(),
                    'status': trs[1].contents[2].text.strip(),
                    'time': int(trs[1].contents[4].text.strip()[:-2]),
                    'memory': int(trs[1].contents[5].text.strip()[:-1]),
                    'ce_info': '',
                }
                if data['status'] == 'Compilation Error':
                    ret, info = self.get_compile_error_info(data['rid'])
                    data['ce_info'] = info if ret else ''
                return True, data
        else:
            return False, '获取结果：http方法错误，请检查网络后重试'

    def get_result(self):
        # 只需要获取自己最近一次提交的结果
        return self.get_result_by_user(self.handle)

    def get_result_by_user(self, handle):
        url = self.url_status + 'user={}'.format(handle)
        return self._get_result(url)

    def get_result_by_rid(self, rid):
        url = self.url_status + 'first={}'.format(rid)
        return self._get_result(url)

    def get_compile_error_info(self, rid):
        url = self.url_home + 'viewerror.php?rid={}'.format(rid)
        ret = self.get(url)
        if ret:
            html = ret.read()
            soup = BeautifulSoup(html, 'html5lib')
            title = soup.find('h1')
            if title.text == 'View Compilation Error':
                return True, soup.find('pre').text
            else:
                return False, soup.find('td', {'valign': 'middle'}).text
        else:
            return False, '获取编译错误信息：http方法错误，请检查网络后重试'
